@@ -9,14 +9,14 @@ class Address(object):
         (?:
             (?P<no>\d+)
             (?P<subno>之\d+)?
-            (?=[巷弄號樓])
+            (?=[巷弄號樓]|$)
             |
             (?P<name>.+?)
         )
         (?:
-            (?P<unit>[縣市鄉鎮市區村里鄰路段街巷弄號樓])
+            (?P<unit>[縣市鄉鎮市區村里鄰路街段巷弄號樓])
             |
-            (?=\d+(?:[之-]\d+)?[巷弄號樓]|$)
+            (?=\d+(?:之\d+)?[巷弄號樓]|$)
         )
     ''', re.X)
 
@@ -26,24 +26,27 @@ class Address(object):
     UNIT  = 3
 
     TO_REPLACE_RE = re.compile(u'''
-    [ 　,，台-]
-    |
-    [０-９]
-    |
-    [一二三四五六七八九]?
-    十?
-    [一二三四五六七八九]
-    (?=[段路街巷弄號樓])
+        [ 　,，台~-]
+        |
+        [０-９]
+        |
+        [一二三四五六七八九]?
+        十?
+        [一二三四五六七八九]
+        (?=[段路街巷弄號樓])
     ''', re.X)
 
-    TO_REMOVE_SET = set(u' 　,，')
-
+    # the strs matched but not in here will be removed
     TO_REPLACE_MAP = {
-        u'-' : u'之', u'台': u'臺',
+        u'-': u'之', u'~': u'之', u'台': u'臺', u'北市': u'臺北市',
+        u'１': u'1', u'２': u'2', u'３': u'3', u'４': u'4', u'５': u'5',
+        u'６': u'6', u'７': u'7', u'８': u'8', u'９': u'9', u'０': u'0',
         u'一': u'1', u'二': u'2', u'三': u'3', u'四': u'4', u'五': u'5',
         u'六': u'6', u'七': u'7', u'八': u'8', u'九': u'9',
     }
 
+    CHINESE_NUMERALS_SET = set(u'一二三四五六七八九十')
+
     @staticmethod
     def normalize(s):
 
@@ -54,49 +57,27 @@ class Address(object):
 
             found = m.group()
 
-            if found in Address.TO_REMOVE_SET:
-                return u''
             if found in Address.TO_REPLACE_MAP:
                 return Address.TO_REPLACE_MAP[found]
 
-            len_found = len(found)
-
-            # 65296 = '０'; 65305 = '９'; 65248 = '０'-'0'
-            if len_found == 1 and 65296 <= ord(found) <= 65305:
-                return unichr(ord(found)-65248)
-
             # for '十一' to '九十九'
-            if len_found == 2:
-                return u'1'+Address.TO_REPLACE_MAP[found[1]]
-            if len_found == 3:
-                return Address.TO_REPLACE_MAP[found[0]]+Address.TO_REPLACE_MAP[found[2]]
+            if found[0] in Address.CHINESE_NUMERALS_SET:
+
+                len_found = len(found)
+                if len_found == 2:
+                    return u'1'+Address.TO_REPLACE_MAP[found[1]]
+                if len_found == 3:
+                    return Address.TO_REPLACE_MAP[found[0]]+Address.TO_REPLACE_MAP[found[2]]
+
+            return u''
 
         s = Address.TO_REPLACE_RE.sub(replace, s)
 
         return s
 
-    TO_INGORE = set(u'鄰里')
-
     @staticmethod
     def tokenize(addr_str):
-
-        tokens = Address.TOKEN_RE.findall(Address.normalize(addr_str))
-
-        # remove tokens whose unit is in TO_INGORE
-
-        len_tokens = len(tokens)
-
-        i = 2
-        while i < len_tokens:
-            name = tokens[i][Address.NAME]
-            unit = tokens[i][Address.UNIT]
-            if unit and unit in Address.TO_INGORE and name != u'宮后':
-                del tokens[i]
-                len_tokens -= 1
-            else:
-                i += 1
-
-        return tuple(tokens)
+        return Address.TOKEN_RE.findall(Address.normalize(addr_str))
 
     def __init__(self, addr_str):
         self.tokens = Address.tokenize(addr_str)
@@ -315,6 +296,7 @@ class Directory(object):
                 )
 
         if len_tokens >= 3:
+            # (a, b, c, d) -> (a, c)
             self.put_gradual(addr.pick_to_flat(0, 2), zipcode)
 
     def within_a_transaction(method):
@@ -382,12 +364,43 @@ class Directory(object):
     def find(self, addr_str):
 
         addr = Address(addr_str)
+        len_addr_tokens = len(addr.tokens)
 
-        for i in range(len(addr.tokens), 0, -1):
+        # avoid unnecessary iteration
+        start_len = len_addr_tokens
+        while start_len >= 0:
+            if addr.parse(start_len-1) == (0, 0):
+                break
+            start_len -= 1
+
+        for i in range(start_len, 0, -1):
 
             addr_str = addr.flat(i)
 
             rzpairs = self.get_rule_str_zipcode_pairs(addr_str)
+
+            # for handling insignificant tokens and redundant unit
+            if (
+                # It only runs once, and must be the first iteration.
+                i == start_len and
+                len_addr_tokens >= 4 and
+                addr.tokens[2][Address.UNIT] in u'村里' and
+                not rzpairs
+            ):
+
+                if addr.tokens[3][Address.UNIT] == u'鄰':
+                    # delete the insignificant token (whose unit is 鄰)
+                    del addr.tokens[3]
+
+                if addr.tokens[3][Address.UNIT] == u'號':
+                    # empty the redundant unit in the token
+                    addr.tokens[2] = (u'', u'', addr.tokens[2][Address.NAME], u'')
+                    rzpairs = self.get_rule_str_zipcode_pairs(addr.flat(3))
+                else:
+                    # delete insignificant token (whose unit is 村 or 里)
+                    del addr.tokens[2]
+                    rzpairs = self.get_rule_str_zipcode_pairs(addr.flat(3))
+
             if rzpairs:
                 for rule_str, zipcode in rzpairs:
                     if Rule(rule_str).match(addr):
